@@ -63,6 +63,7 @@ enum cmd_and_opt_values
   
   oNoVerbose = 500,
   aGPGConfList,
+  aGPGConfTest,
   oOptions,
   oDebug,
   oDebugAll,
@@ -97,6 +98,7 @@ enum cmd_and_opt_values
 static ARGPARSE_OPTS opts[] = {
 
   { aGPGConfList, "gpgconf-list", 256, "@" },
+  { aGPGConfTest, "gpgconf-test", 256, "@" },
   
   { 301, NULL, 0, N_("@Options:\n ") },
 
@@ -137,9 +139,11 @@ static ARGPARSE_OPTS opts[] = {
 };
 
 
-/* The card dirver we use by default for PC/SC.  */
+/* The card driver we use by default for PC/SC.  */
 #if defined(HAVE_W32_SYSTEM) || defined(__CYGWIN__)
 #define DEFAULT_PCSC_DRIVER "winscard.dll"
+#elif defined(__APPLE__)
+#define DEFAULT_PCSC_DRIVER "/System/Library/Frameworks/PCSC.framework/PCSC"
 #elif defined(__GLIBC__)
 #define DEFAULT_PCSC_DRIVER "libpcsclite.so.1"
 #else
@@ -217,25 +221,6 @@ i18n_init (void)
 
 
 
-/* Used by gcry for logging */
-static void
-my_gcry_logger (void *dummy, int level, const char *fmt, va_list arg_ptr)
-{
-  /* translate the log levels */
-  switch (level)
-    {
-    case GCRY_LOG_CONT: level = JNLIB_LOG_CONT; break;
-    case GCRY_LOG_INFO: level = JNLIB_LOG_INFO; break;
-    case GCRY_LOG_WARN: level = JNLIB_LOG_WARN; break;
-    case GCRY_LOG_ERROR:level = JNLIB_LOG_ERROR; break;
-    case GCRY_LOG_FATAL:level = JNLIB_LOG_FATAL; break;
-    case GCRY_LOG_BUG:  level = JNLIB_LOG_BUG; break;
-    case GCRY_LOG_DEBUG:level = JNLIB_LOG_DEBUG; break;
-    default:            level = JNLIB_LOG_ERROR; break;  
-    }
-  log_logv (level, fmt, arg_ptr);
-}
-
 
 /* Setup the debugging.  With a LEVEL of NULL only the active debug
    flags are propagated to the subsystems.  With LEVEL set, a specific
@@ -351,7 +336,7 @@ main (int argc, char **argv )
      the option parsing may need services of the library */
   if (!gcry_check_version (NEED_LIBGCRYPT_VERSION) )
     {
-      log_fatal( _("libgcrypt is too old (need %s, have %s)\n"),
+      log_fatal (_("%s is too old (need %s, have %s)\n"), "libgcrypt",
                  NEED_LIBGCRYPT_VERSION, gcry_check_version (NULL) );
     }
 
@@ -360,8 +345,9 @@ main (int argc, char **argv )
   assuan_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free);
   assuan_set_assuan_log_stream (log_get_stream ());
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
+  assuan_set_assuan_err_source (GPG_ERR_SOURCE_DEFAULT);
 
-  gcry_set_log_handler (my_gcry_logger, NULL);
+  setup_libgcrypt_logging ();
   gcry_control (GCRYCTL_USE_SECURE_RNDPOOL);
 
   may_coredump = disable_core_dumps ();
@@ -448,6 +434,7 @@ main (int argc, char **argv )
       switch (pargs.r_opt)
         {
         case aGPGConfList: gpgconf_list = 1; break;
+        case aGPGConfTest: gpgconf_list = 2; break;
         case oQuiet: opt.quiet = 1; break;
         case oVerbose: opt.verbose++; break;
         case oBatch: opt.batch=1; break;
@@ -550,6 +537,8 @@ main (int argc, char **argv )
       log_debug ("... okay\n");
     }
   
+  if (gpgconf_list == 2)
+    scd_exit (0);
   if (gpgconf_list)
     {
       /* List options and default values in the GPG Conf format.  */
@@ -593,7 +582,7 @@ main (int argc, char **argv )
       printf ("disable-ccid:%lu:\n", GC_OPT_FLAG_NONE );
 #endif
       printf ("allow-admin:%lu:\n", GC_OPT_FLAG_NONE );
-
+      printf ("disable-keypad:%lu:\n", GC_OPT_FLAG_NONE );
 
       scd_exit (0);
     }
@@ -608,6 +597,7 @@ main (int argc, char **argv )
   if (pipe_server)
     { 
       /* This is the simple pipe based server */
+      ctrl_t ctrl;
       pth_attr_t tattr;
       int fd = -1;
 
@@ -649,10 +639,19 @@ main (int argc, char **argv )
       pth_attr_set (tattr, PTH_ATTR_STACK_SIZE, 512*1024);
       pth_attr_set (tattr, PTH_ATTR_NAME, "pipe-connection");
 
-      if (!pth_spawn (tattr, start_connection_thread, (void*)(-1)))
+      ctrl = xtrycalloc (1, sizeof *ctrl);
+      if ( !ctrl )
+        {
+          log_error ("error allocating connection control data: %s\n",
+                     strerror (errno) );
+          scd_exit (2);
+        }
+      ctrl->thread_startup.fd = -1;
+      if ( !pth_spawn (tattr, start_connection_thread, ctrl) )
         {
           log_error ("error spawning pipe connection handler: %s\n",
                      strerror (errno) );
+          xfree (ctrl);
           scd_exit (2);
         }
 
@@ -803,12 +802,18 @@ scd_exit (int rc)
 }
 
 
-void
+static void
 scd_init_default_ctrl (ctrl_t ctrl)
 {
   ctrl->reader_slot = -1;
 }
 
+static void
+scd_deinit_default_ctrl (ctrl_t ctrl)
+{
+
+}
+
 
 /* Return the name of the socket to be used to connect to this
    process.  If no socket is available, return NULL. */
@@ -1000,23 +1005,26 @@ create_server_socket (int is_standard_name, const char *name)
 static void *
 start_connection_thread (void *arg)
 {
-  int fd = (int)arg;
+  ctrl_t ctrl = arg;
+
+  scd_init_default_ctrl (ctrl);
+  if (opt.verbose)
+    log_info (_("handler for fd %d started\n"), ctrl->thread_startup.fd);
+
+  scd_command_handler (ctrl, ctrl->thread_startup.fd);
 
   if (opt.verbose)
-    log_info (_("handler for fd %d started\n"), fd);
-
-  scd_command_handler (fd);
-
-  if (opt.verbose)
-    log_info (_("handler for fd %d terminated\n"), fd);
+    log_info (_("handler for fd %d terminated\n"), ctrl->thread_startup.fd);
 
   /* If this thread is the pipe connection thread, flag that a
      shutdown is required.  With the next ticker event and given that
      no other connections are running the shutdown will then
      happen. */
-  if (fd == -1)
+  if (ctrl->thread_startup.fd == -1)
     shutdown_pending = 1;
   
+  scd_deinit_default_ctrl (ctrl);
+  xfree (ctrl);
   return NULL;
 }
 
@@ -1130,23 +1138,33 @@ handle_connections (int listen_fd)
 
       if (listen_fd != -1 && FD_ISSET (listen_fd, &read_fdset))
 	{
+          ctrl_t ctrl;
+
           plen = sizeof paddr;
 	  fd = pth_accept (listen_fd, (struct sockaddr *)&paddr, &plen);
 	  if (fd == -1)
 	    {
 	      log_error ("accept failed: %s\n", strerror (errno));
 	    }
+          else if ( !(ctrl = xtrycalloc (1, sizeof *ctrl)) )
+            {
+              log_error ("error allocating connection control data: %s\n",
+                         strerror (errno) );
+              close (fd);
+            }
           else
             {
               char threadname[50];
+
               snprintf (threadname, sizeof threadname-1, "conn fd=%d", fd);
               threadname[sizeof threadname -1] = 0;
               pth_attr_set (tattr, PTH_ATTR_NAME, threadname);
-              
-              if (!pth_spawn (tattr, start_connection_thread, (void*)fd))
+              ctrl->thread_startup.fd = fd;
+              if (!pth_spawn (tattr, start_connection_thread, ctrl))
                 {
                   log_error ("error spawning connection handler: %s\n",
                              strerror (errno) );
+                  xfree (ctrl);
                   close (fd);
                 }
             }
