@@ -1,5 +1,5 @@
 /* gpg-connect-agent.c - Tool to connect to the agent.
- *	Copyright (C) 2005 Free Software Foundation, Inc.
+ *	Copyright (C) 2005, 2007 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -41,10 +41,13 @@ enum cmd_and_opt_values
     oQuiet      = 'q',
     oVerbose	= 'v',
     oRawSocket  = 'S',
+    oExec       = 'E',
 
     oNoVerbose	= 500,
     oHomedir,
-    oHex
+    oHex,
+    oDecode,
+    oNoExtConnect
 
   };
 
@@ -57,7 +60,11 @@ static ARGPARSE_OPTS opts[] =
     { oVerbose, "verbose",  0, N_("verbose") },
     { oQuiet, "quiet",      0, N_("quiet") },
     { oHex,   "hex",        0, N_("print data out hex encoded") },
+    { oDecode,"decode",     0, N_("decode received data lines") },
     { oRawSocket, "raw-socket", 2, N_("|NAME|connect to Assuan socket NAME")},
+    { oExec, "exec", 0, N_("run the Assuan server given on the command line")},
+    { oNoExtConnect, "no-ext-connect",
+                            0, N_("do not use extended connect mode")},
 
     /* hidden options */
     { oNoVerbose, "no-verbose",  0, "@"},
@@ -73,7 +80,10 @@ struct
   int quiet;		/* Be extra quiet.  */
   const char *homedir;  /* Configuration directory name */
   int hex;              /* Print data lines in hex format. */
+  int decode;           /* Decode received data lines.  */
   const char *raw_socket; /* Name of socket to connect in raw mode. */
+  int exec;             /* Run the pgm given on the command line. */
+  unsigned int connect_flags;    /* Flags used for connecting. */
 } opt;
 
 
@@ -209,13 +219,68 @@ clear_definq (void)
 }      
 
 
+static void
+do_sendfd (assuan_context_t ctx, char *line)
+{
+  FILE *fp;
+  char *name, *mode, *p;
+  int rc, fd;
+
+  /* Get file name. */
+  name = line;
+  for (p=name; *p && !spacep (p); p++)
+    ;
+  if (*p)
+    *p++ = 0;
+  while (spacep (p))
+    p++;
+
+  /* Get mode.  */
+  mode = p;
+  if (!*mode)
+    mode = "r";
+  else
+    {
+      for (p=mode; *p && !spacep (p); p++)
+        ;
+      if (*p)
+        *p++ = 0;
+    }
+
+  /* Open and send. */
+  fp = fopen (name, mode);
+  if (!fp)
+    {
+      log_error ("can't open `%s' in \"%s\" mode: %s\n",
+                 name, mode, strerror (errno));
+      return;
+    }
+  fd = fileno (fp);
+
+  if (opt.verbose)
+    log_error ("file `%s' opened in \"%s\" mode, fd=%d\n",
+               name, mode, fd);
+
+  rc = assuan_sendfd (ctx, fd);
+  if (rc)
+    log_error ("sednig  descriptor %d failed: %s\n", fd, gpg_strerror (rc));
+  fclose (fp);
+}
+
+
+static void
+do_recvfd (assuan_context_t ctx, char *line)
+{
+  log_info ("This command has not yet been implemented\n");
+}
+
+
 
 /* gpg-connect-agent's entry point. */
 int
 main (int argc, char **argv)
 {
   ARGPARSE_ARGS pargs;
-  const char *fname;
   int no_more_options = 0;
   assuan_context_t ctx;
   char *line, *p;
@@ -224,10 +289,12 @@ main (int argc, char **argv)
 
   set_strusage (my_strusage);
   log_set_prefix ("gpg-connect-agent", 1);
+  assuan_set_assuan_err_source (0);
 
   i18n_init();
 
   opt.homedir = default_homedir ();
+  opt.connect_flags = 1; /* Use extended connect mode.  */
 
   /* Parse the command line. */
   pargs.argc  = &argc;
@@ -242,7 +309,10 @@ main (int argc, char **argv)
         case oNoVerbose: opt.verbose = 0; break;
         case oHomedir:   opt.homedir = pargs.r.ret_str; break;
         case oHex:       opt.hex = 1; break;
+        case oDecode:    opt.decode = 1; break;
         case oRawSocket: opt.raw_socket = pargs.r.ret_str; break;
+        case oExec:      opt.exec = 1; break;
+        case oNoExtConnect: opt.connect_flags &= ~(1); break;
 
         default: pargs.err = 2; break;
 	}
@@ -250,16 +320,52 @@ main (int argc, char **argv)
 
   if (log_get_errorcount (0))
     exit (2);
-  
-  fname = argc ? *argv : NULL;
 
-  if (opt.raw_socket)
+  if (opt.exec)
     {
-      rc = assuan_socket_connect (&ctx, opt.raw_socket, 0);
+      if (!argc)
+        {
+          log_error (_("option \"%s\" requires a program "
+                       "and optional arguments\n"), "--exec" );
+          exit (1);
+        }
+    }
+  else if (argc)
+    usage (1);
+
+  if (opt.exec && opt.raw_socket)
+    log_info (_("option \"%s\" ignored due to \"%s\"\n"),
+              "--raw-socket", "--exec");
+
+  if (opt.exec)
+    {
+      int no_close[3];
+
+      no_close[0] = fileno (stderr);
+      no_close[1] = log_get_fd ();
+      no_close[2] = -1;
+      rc = assuan_pipe_connect_ext (&ctx, *argv, (const char **)argv,
+                                    no_close, NULL, NULL,
+                                    opt.connect_flags);
+      if (rc)
+        {
+          log_error ("assuan_pipe_connect_ext failed: %s\n",
+                     gpg_strerror (rc));
+          exit (1);
+        }
+
+      if (opt.verbose)
+        log_info ("server `%s' started\n", *argv);
+
+    }
+  else if (opt.raw_socket)
+    {
+      rc = assuan_socket_connect_ext (&ctx, opt.raw_socket, 0,
+                                      opt.connect_flags);
       if (rc)
         {
           log_error ("can't connect to socket `%s': %s\n",
-                     opt.raw_socket, assuan_strerror (rc));
+                     opt.raw_socket, gpg_strerror (rc));
           exit (1);
         }
 
@@ -324,19 +430,42 @@ main (int argc, char **argv)
             {
               puts (p);
             }
+          else if (!strcmp (cmd, "sendfd"))
+            {
+              do_sendfd (ctx, p);
+              continue;
+            }
+          else if (!strcmp (cmd, "recvfd"))
+            {
+              do_recvfd (ctx, p);
+              continue;
+            }
+          else if (!strcmp (cmd, "hex"))
+            opt.hex = 1;
+          else if (!strcmp (cmd, "nohex"))
+            opt.hex = 0;
+          else if (!strcmp (cmd, "decode"))
+            opt.decode = 1;
+          else if (!strcmp (cmd, "nodecode"))
+            opt.decode = 0;
           else if (!strcmp (cmd, "help"))
             {
-              puts ("Available commands:\n"
-                    "/echo ARGS             Echo ARGS.\n"
-                    "/definqfile NAME FILE\n"
-                    "    Use content of FILE for inquiries with NAME.\n"
-                    "    NAME may be \"*\" to match any inquiry.\n"
-                    "/definqprog NAME PGM\n"
-                    "    Run PGM for inquiries matching NAME and pass the\n"
-                    "    entire line to it as arguments.\n"
-                    "/showdef               Print all definitions.\n"
-                    "/cleardef              Delete all definitions.\n"
-                    "/help                  Print this help.");
+              puts (
+"Available commands:\n"
+"/echo ARGS             Echo ARGS.\n"
+"/definqfile NAME FILE\n"
+"    Use content of FILE for inquiries with NAME.\n"
+"    NAME may be \"*\" to match any inquiry.\n"
+"/definqprog NAME PGM\n"
+"    Run PGM for inquiries matching NAME and pass the\n"
+"    entire line to it as arguments.\n"
+"/showdef               Print all definitions.\n"
+"/cleardef              Delete all definitions.\n"
+"/sendfd FILE MODE      Open FILE and pass descriptor to server.\n"
+"/recvfd                Receive FD from server and print. \n"
+"/[no]hex               Enable hex dumping of received data lines.\n"
+"/[no]decode            Enable decoding of received data lines.\n"
+"/help                  Print this help.");
             }
           else
             log_error (_("unknown command `%s'\n"), cmd );
@@ -347,15 +476,15 @@ main (int argc, char **argv)
       rc = assuan_write_line (ctx, line);
       if (rc)
         {
-          log_info (_("sending line failed: %s\n"), assuan_strerror (rc) );
+          log_info (_("sending line failed: %s\n"), gpg_strerror (rc) );
           continue;
         }
       if (*line == '#' || !*line)
-        continue; /* Don't expect a response for a coment line. */
+        continue; /* Don't expect a response for a comment line. */
 
       rc = read_and_print_response (ctx);
       if (rc)
-        log_info (_("receiving line failed: %s\n"), assuan_strerror (rc) );
+        log_info (_("receiving line failed: %s\n"), gpg_strerror (rc) );
     }
 
   if (opt.verbose)
@@ -431,7 +560,7 @@ handle_inquire (assuan_context_t ctx, char *line)
       rc = assuan_send_data (ctx, buffer, n);
       if (rc)
         {
-          log_error ("sending data back failed: %s\n", assuan_strerror (rc) );
+          log_error ("sending data back failed: %s\n", gpg_strerror (rc) );
           break;
         }
     }
@@ -440,7 +569,7 @@ handle_inquire (assuan_context_t ctx, char *line)
 
   rc = assuan_send_data (ctx, NULL, 0);
   if (rc)
-    log_error ("sending data back failed: %s\n", assuan_strerror (rc) );
+    log_error ("sending data back failed: %s\n", gpg_strerror (rc) );
 
   if (d->is_prog)
     {
@@ -462,6 +591,7 @@ read_and_print_response (assuan_context_t ctx)
   size_t linelen;
   assuan_error_t rc;
   int i, j;
+  int need_lf = 0;
 
   for (;;)
     {
@@ -470,6 +600,12 @@ read_and_print_response (assuan_context_t ctx)
           rc = assuan_read_line (ctx, &line, &linelen);
           if (rc)
             return rc;
+
+          if (opt.verbose > 1 && *line == '#')
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+            }
         }    
       while (*line == '#' || !linelen);
 
@@ -507,56 +643,92 @@ read_and_print_response (assuan_context_t ctx)
                   putchar ('\n');
                 }
             }
+          else if (opt.decode)
+            {
+              const unsigned char *s;
+              int need_d = 1;
+              int c = 0;
+
+              for (j=2, s=(unsigned char*)line+2; j < linelen; j++, s++ )
+                {
+                  if (need_d)
+                    {
+                      fputs ("D ", stdout);
+                      need_d = 0;
+                    }
+                  if (*s == '%' && j+2 < linelen)
+                    { 
+                      s++; j++;
+                      c = xtoi_2 ( s );
+                      s++; j++;
+                    }
+                  else
+                    c = *s;
+                  if (c == '\n')
+                    need_d = 1;
+                  putchar (c);
+                }
+              need_lf = (c != '\n');
+            }
           else
             {
               fwrite (line, linelen, 1, stdout);
               putchar ('\n');
             }
         }
-      else if (linelen >= 1
-               && line[0] == 'S' 
-               && (line[1] == '\0' || line[1] == ' '))
+      else 
         {
-          fwrite (line, linelen, 1, stdout);
-          putchar ('\n');
-        }  
-      else if (linelen >= 2
-               && line[0] == 'O' && line[1] == 'K'
-               && (line[2] == '\0' || line[2] == ' '))
-        {
-          fwrite (line, linelen, 1, stdout);
-          putchar ('\n');
-          return 0;
+          if (need_lf)
+            {
+              putchar ('\n');
+              need_lf = 0;
+            }
+
+          if (linelen >= 1
+              && line[0] == 'S' 
+              && (line[1] == '\0' || line[1] == ' '))
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+            }  
+          else if (linelen >= 2
+                   && line[0] == 'O' && line[1] == 'K'
+                   && (line[2] == '\0' || line[2] == ' '))
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+              return 0;
+            }
+          else if (linelen >= 3
+                   && line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
+                   && (line[3] == '\0' || line[3] == ' '))
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+              return 0;
+            }  
+          else if (linelen >= 7
+                   && line[0] == 'I' && line[1] == 'N' && line[2] == 'Q'
+                   && line[3] == 'U' && line[4] == 'I' && line[5] == 'R'
+                   && line[6] == 'E' 
+                   && (line[7] == '\0' || line[7] == ' '))
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+              if (!handle_inquire (ctx, line))
+                assuan_write_line (ctx, "CANCEL");
+            }
+          else if (linelen >= 3
+                   && line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
+                   && (line[3] == '\0' || line[3] == ' '))
+            {
+              fwrite (line, linelen, 1, stdout);
+              putchar ('\n');
+              /* Received from server, thus more responses are expected.  */
+            }
+          else
+            return gpg_error (GPG_ERR_ASS_INV_RESPONSE);
         }
-      else if (linelen >= 3
-               && line[0] == 'E' && line[1] == 'R' && line[2] == 'R'
-               && (line[3] == '\0' || line[3] == ' '))
-        {
-          fwrite (line, linelen, 1, stdout);
-          putchar ('\n');
-          return 0;
-        }  
-      else if (linelen >= 7
-               && line[0] == 'I' && line[1] == 'N' && line[2] == 'Q'
-               && line[3] == 'U' && line[4] == 'I' && line[5] == 'R'
-               && line[6] == 'E' 
-               && (line[7] == '\0' || line[7] == ' '))
-        {
-          fwrite (line, linelen, 1, stdout);
-          putchar ('\n');
-          if (!handle_inquire (ctx, line))
-            assuan_write_line (ctx, "CANCEL");
-        }
-      else if (linelen >= 3
-               && line[0] == 'E' && line[1] == 'N' && line[2] == 'D'
-               && (line[3] == '\0' || line[3] == ' '))
-        {
-          fwrite (line, linelen, 1, stdout);
-          putchar ('\n');
-          /* Received from server, thus more responses are expected.  */
-        }
-      else
-        return ASSUAN_Invalid_Response;
     }
 }
 
@@ -612,7 +784,7 @@ start_agent (void)
 
   if (rc)
     {
-      log_error ("can't connect to the agent: %s\n", assuan_strerror (rc));
+      log_error ("can't connect to the agent: %s\n", gpg_strerror (rc));
       exit (1);
     }
 
@@ -623,7 +795,7 @@ start_agent (void)
   if (rc)
     {
       log_error (_("error sending %s command: %s\n"), "RESET", 
-                 assuan_strerror (rc));
+                 gpg_strerror (rc));
       exit (1);
     }
 
